module Backup.Directories
    (
      allPaths
    , getMtimes
    , getAllItems
    , getAllMtimes
    , buildMap
    , toBackup
    , readMap
    , writeMap
    , BackupDir(..)
    , Metadata(..)
    , FileMetadata
    ) where

import Backup.Sorting
import Control.Exception
import Data.Binary.Put
import Data.Map
import System.FilePath
import System.Directory
import System.Posix.Files
import System.Posix.Types


data Metadata = Metadata { base  :: FilePath
                         , isDir :: Bool
                         , mtime :: EpochTime
                         } deriving (Eq, Show, Read)
instance Ord Metadata where
    m1 <= m2 = (mtime m1) <= (mtime m2)
data BackupDir = BackupDir { basePath :: FilePath
                           , dirs :: [FilePath]
                           } deriving (Show)
type FileMetadata = (FilePath, Metadata)
type DirMap = Map FilePath Metadata


getAllItems :: [FilePath] -> IO [BackupDir]
getAllItems [] = return []
getAllItems (x:xs) = do
    this <- allPaths x
    rest <- getAllItems xs
    return $ (BackupDir {basePath = "./", dirs = this}) : rest


getAllMtimes :: [BackupDir] -> IO [FileMetadata]
getAllMtimes [] = return []
getAllMtimes (x:xs) = do
    this <- getMtimes (basePath x) (dirs x)
    rest <- getAllMtimes xs
    return $ this ++ rest


allPaths :: FilePath -> IO [FilePath]
allPaths root = do
    status <- try (getFileStatus root) :: IO (Either SomeException FileStatus)
    case status of
        Left ex -> putStrLn (show ex) >> return []
        Right s -> if isDirectory s then recurse root else return [root]


recurse :: FilePath -> IO [FilePath]
recurse path = do
    children <- try (listDirectory path) :: IO (Either SomeException [FilePath])
    items <- case children of
        Left ex -> putStrLn (show ex) >> return []
        Right s -> sequence (allPaths <$> ((</>) path <$> s))
    return $ concat items


getMtimes :: FilePath -> [FilePath] -> IO [FileMetadata]
getMtimes _ [] = return []
getMtimes base (x:xs) = do
    file <- getFileStatus x
    rest <- getMtimes base xs
    return $ (x, Metadata base (isDirectory file) (modificationTime file)) : rest


buildMap :: [FileMetadata] -> DirMap
buildMap = fromList


toBackup :: DirMap -> DirMap -> [FileMetadata]
toBackup records current = sort (diff ++ unio)
  where
    diff = toList (difference current records)
    unio = [(k, m) | (k, (m, v)) <- toList intersect, v]
    intersect = intersectionWith combiner records current
    combiner :: Metadata -> Metadata -> (Metadata, Bool)
    combiner e1 e2 = if b1 == b2 && m1 == m2 then (e2, False) else (e2, True)
      where
        b1 = base e1
        b2 = base e2
        m1 = mtime e1
        m2 = mtime e2


readMap :: IO DirMap
readMap = do
    str <- readFile "records.txt"
    return (read str :: DirMap)

writeMap :: DirMap -> IO ()
writeMap m = do
    let l = toList m
    let str = show m
    writeFile "records.txt" str

